@@ -1,7 +1,7 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
  * Copyright (c) 2007 Mario Danic
- * Copyright (c) 2009 - 2012 Thomas Schmitt
+ * Copyright (c) 2009 - 2020 Thomas Schmitt
  * 
  * This file is part of the libisofs project; you can redistribute it and/or 
  * modify it under the terms of the GNU General Public License version 2 
@@ -32,10 +32,16 @@
 #define ISO_ROCKRIDGE_IN_DIR_REC 124
 #endif
 
+#define ISO_CE_ENTRY_SIZE  28
+
 
 static
 int susp_add_ES(Ecma119Image *t, struct susp_info *susp, int to_ce, int seqno);
 
+static
+int susp_make_CE(Ecma119Image *t, uint8_t **CE,
+                 uint32_t block_offset, uint32_t byte_offset, uint32_t size);
+
 
 static
 int susp_append(Ecma119Image *t, struct susp_info *susp, uint8_t *data)
@@ -54,14 +60,76 @@ int susp_append(Ecma119Image *t, struct susp_info *susp, uint8_t *data)
 static
 int susp_append_ce(Ecma119Image *t, struct susp_info *susp, uint8_t *data)
 {
+    int to_alloc = 1, ret;
+    unsigned char *pad;
+    uint8_t *CE;
+    size_t next_alloc;
+
+    if (data[0] &&
+        (susp->ce_len + data[2] + ISO_CE_ENTRY_SIZE - 1) / BLOCK_SIZE !=
+                                                   susp->ce_len / BLOCK_SIZE) {
+        /* Linux fs/isofs wants byte_offset + ce_len <= BLOCK_SIZE.
+           So this Continuation Area needs to end by a CE which points
+           to the start of the next block.
+        */ 
+        to_alloc = 2;
+        if ((susp->ce_len + ISO_CE_ENTRY_SIZE) % BLOCK_SIZE)
+            to_alloc = 3; /* need a PAD pseudo entry */
+    }
+
+    if (susp->ce_susp_fields == NULL)
+        susp->alloc_ce_susp_fields = 0;
+    if (susp->n_ce_susp_fields + to_alloc > susp->alloc_ce_susp_fields) {
+        next_alloc = susp->alloc_ce_susp_fields;
+        while (susp->n_ce_susp_fields + to_alloc > next_alloc)
+            next_alloc += ISO_SUSP_CE_ALLOC_STEP;
+        susp->ce_susp_fields = realloc(susp->ce_susp_fields,
+                                       sizeof(uint8_t *) * next_alloc);
+        if (susp->ce_susp_fields == NULL)
+            return ISO_OUT_OF_MEM;
+        susp->alloc_ce_susp_fields = next_alloc;
+    }
+
+    if (to_alloc >= 2) {
+        /* Insert CE entry (actual CE size later by susp_update_CE_sizes) */
+        ret = susp_make_CE(t, &CE, (uint32_t) (susp->ce_block +
+                                               susp->ce_len / BLOCK_SIZE + 1),
+                           (uint32_t) 0, (uint32_t) 2048);
+        if (ret < 0)
+            return ret;
+        susp->ce_susp_fields[susp->n_ce_susp_fields] = CE;
+        susp->ce_len += ISO_CE_ENTRY_SIZE;
+        susp->n_ce_susp_fields++;
+    }
+    if (to_alloc >= 3) {
+
+
+#ifdef Libisofs_ce_calc_debuG
+
+        fprintf(stderr,
+                  "\nlibburn_DEBUG: Inserting %d bytes of CE padding\n\n",
+                  (int) (BLOCK_SIZE - (susp->ce_len % BLOCK_SIZE)));
+
+#endif /* Libisofs_ce_calc_debuG */
+
+        pad = malloc(1);
+        if (pad == NULL)
+            return ISO_OUT_OF_MEM;
+        pad[0] = 0;
+        susp->ce_susp_fields[susp->n_ce_susp_fields] = pad;
+        if (susp->ce_len % BLOCK_SIZE)
+            susp->ce_len += BLOCK_SIZE - (susp->ce_len % BLOCK_SIZE);
+        susp->n_ce_susp_fields++;
+    }
+    susp->ce_susp_fields[susp->n_ce_susp_fields] = data;
     susp->n_ce_susp_fields++;
-    susp->ce_susp_fields = realloc(susp->ce_susp_fields, sizeof(void*)
-                                   * susp->n_ce_susp_fields);
-    if (susp->ce_susp_fields == NULL) {
-        return ISO_OUT_OF_MEM;
+
+    if (data[0] == 0) {
+        if (susp->ce_len % BLOCK_SIZE)
+            susp->ce_len += BLOCK_SIZE - (susp->ce_len % BLOCK_SIZE);
+    } else {
+        susp->ce_len += data[2];
     }
-    susp->ce_susp_fields[susp->n_ce_susp_fields - 1] = data;
-    susp->ce_len += data[2];
     return ISO_SUCCESS;
 }
 
@@ -310,14 +378,12 @@ int iso_get_rr_name(IsoWriteOpts *opts, char *input_charset,
 
     ret = strconv(str, input_charset, output_charset, name);
     if (ret < 0) {
-        /* TODO we should check for possible cancelation */
+        /* TODO we should check for possible cancellation */
         if (!(flag & 1))
             iso_msg_submit(imgid, ISO_FILENAME_WRONG_CHARSET, ret,
                    "Charset conversion error. Cannot convert %s from %s to %s",
                    str, input_charset, output_charset);
-
-        /* use the original name, it's the best we can do */
-        ret = iso_clone_mem(str, name, 0);
+        *name = NULL;
         return ISO_FILENAME_WRONG_CHARSET;
     }
 
@@ -507,16 +573,19 @@ int rrip_add_SL(Ecma119Image *t, struct susp_info *susp, uint8_t **comp,
             }
 
             /*
-             * In this case we are sure we're writting to CE. Check for
+             * In this case we are sure we're writing to CE. Check for
              * debug purposes
              */
             if (ce == 0) {
+                free(SL);
                 return ISO_ASSERT_FAILURE;
             }
             ret = susp_append_ce(t, susp, SL);
             if (ret < 0) {
+                free(SL);
                 return ret;
             }
+            SL = NULL; /* now owned by susp */
             written = i;
             total_comp_len = comp[i][1] + 2;
         }
@@ -547,22 +616,56 @@ int rrip_add_SL(Ecma119Image *t, struct susp_info *susp, uint8_t **comp,
 }
 
 
+/* @param flag bit1= care about crossing block boundaries */
+static
+int susp_calc_add_to_ce(size_t *ce, size_t base_ce, int add, int flag)
+{
+    if (flag & 2) {
+        /* Account for inserted CE before size exceeds block size */
+        if ((*ce + base_ce + add + ISO_CE_ENTRY_SIZE - 1) / BLOCK_SIZE !=
+            (*ce + base_ce) / BLOCK_SIZE) {
+            /* Insert CE and padding */
+            *ce += ISO_CE_ENTRY_SIZE;
+            if ((*ce + base_ce) % BLOCK_SIZE)
+                *ce += BLOCK_SIZE - ((*ce + base_ce) % BLOCK_SIZE);
+        }
+    }
+    *ce += add;
+    return ISO_SUCCESS;
+}
+
+
 /*
-   @param flag bit0= only account sizes in sua_free resp. ce_len
-                     parameters susp and data may be NULL in this case
+   @param flag bit0= only account sizes in sua_free resp. ce_len.
+                     Parameter susp may be NULL in this case
+               bit1= account for crossing block boundaries
+                     (implied by bit0 == 0)
+   @param ce_len counts the freshly added CA size of the current node
+   @param ce_mem tells the CA size of previous nodes in the same directory
 */
 static
 int aaip_add_AL(Ecma119Image *t, struct susp_info *susp,
                 uint8_t **data, size_t num_data,
-                size_t *sua_free, size_t *ce_len, int flag)
+                size_t *sua_free, size_t *ce_len, size_t ce_mem, int flag)
 {
     int ret, done = 0, len, es_extra = 0;
     uint8_t *aapt, *cpt;
+    size_t count = 0;
 
+    if (!(flag & 1))
+        flag |= 2;
     if (!t->opts->aaip_susp_1_10)
         es_extra = 5;
     if (*sua_free < num_data + es_extra || *ce_len > 0) {
-        *ce_len += num_data + es_extra;
+        if (es_extra > 0)
+            susp_calc_add_to_ce(ce_len, ce_mem, es_extra, flag & 2);
+        done = 0;
+        for (aapt = *data; !done; aapt += aapt[2]) {
+            done = !(aapt[4] & 1);
+            len = aapt[2];
+            susp_calc_add_to_ce(ce_len, ce_mem, len, flag & 2);
+            count += len;
+        }
     } else {
         *sua_free -= num_data + es_extra;
     }
@@ -588,6 +691,7 @@ int aaip_add_AL(Ecma119Image *t, struct susp_info *susp,
     }
 
     /* Multiple fields have to be handed over as single field copies */
+    done = 0;
     for (aapt = *data; !done; aapt += aapt[2]) {
         done = !(aapt[4] & 1);
         len = aapt[2];
@@ -722,6 +826,35 @@ int aaip_add_ER(Ecma119Image *t, struct susp_info *susp, int flag)
 }
 
 
+/**
+ * Create the byte representation of a CE entry.
+ * (SUSP, 5.1).
+ */
+static
+int susp_make_CE(Ecma119Image *t, uint8_t **CE,
+                 uint32_t block_offset, uint32_t byte_offset, uint32_t size)
+{
+    uint8_t *data;
+
+    *CE = NULL;
+    data = calloc(1, 28);
+    if (data == NULL)
+        return ISO_OUT_OF_MEM;
+    *CE = data;
+
+    data[0] = 'C';
+    data[1] = 'E';
+    data[2] = 28;
+    data[3] = 1;
+
+    iso_bb(&data[4], block_offset - t->eff_partition_offset, 4);
+    iso_bb(&data[12], byte_offset, 4);
+    iso_bb(&data[20], size, 4);
+
+    return ISO_SUCCESS;
+}
+
+
 /**
  * Add a CE System Use Entry to the given tree node. A "CE" is used to add
  * a continuation area, where additional System Use Entry can be written.
@@ -730,20 +863,18 @@ int aaip_add_ER(Ecma119Image *t, struct susp_info *susp, int flag)
 static
 int susp_add_CE(Ecma119Image *t, size_t ce_len, struct susp_info *susp)
 {
-    uint8_t *CE = malloc(28);
-    if (CE == NULL) {
-        return ISO_OUT_OF_MEM;
-    }
-
-    CE[0] = 'C';
-    CE[1] = 'E';
-    CE[2] = 28;
-    CE[3] = 1;
-
-    iso_bb(&CE[4], susp->ce_block - t->eff_partition_offset, 4);
-    iso_bb(&CE[12], susp->ce_len, 4);
-    iso_bb(&CE[20], (uint32_t) ce_len, 4);
+    uint32_t block_offset, byte_offset;
+    uint8_t *CE;
+    int ret;
 
+    /* Linux fs/isofs wants byte_offset + ce_len <= BLOCK_SIZE
+     * Here the byte_offset is reduced to the minimum.
+     */
+    block_offset = susp->ce_block + susp->ce_len / BLOCK_SIZE;
+    byte_offset = susp->ce_len % BLOCK_SIZE;
+    ret = susp_make_CE(t, &CE, block_offset, byte_offset, (uint32_t) ce_len);
+    if (ret < 0)
+        return ret;
     return susp_append(t, susp, CE);
 }
 
@@ -799,28 +930,65 @@ int susp_add_ES(Ecma119Image *t, struct susp_info *susp, int to_ce, int seqno)
 }
 
 
+/** 
+ * A field beginning by 0 causes rrip_write_ce_fields() to pad up to the
+ * next block.
+ */
+static
+int pseudo_susp_add_PAD(Ecma119Image *t, struct susp_info *susp)
+{
+    unsigned char *pad;
+    int ret;
+
+    pad = malloc(1);
+    if (pad == NULL)
+        return ISO_OUT_OF_MEM;
+    pad[0] = 0;
+    ret = susp_append_ce(t, susp, pad);
+    if (ret < 0)
+        return ret;
+    return ISO_SUCCESS;
+}
+
+
 /**
  * see doc/zisofs_format.txt : "ZF System Use Entry Format"
+ * see doc/zisofs2_format.txt : "ZF System Use Entry Format", "Z2 ..."
  */
 static
 int zisofs_add_ZF(Ecma119Image *t, struct susp_info *susp, int to_ce,
-                  int header_size_div4, int block_size_log2,
-                  uint32_t uncompressed_size, int flag)
+                  uint8_t algo[2], int header_size_div4, int block_size_log2,
+                  uint64_t uncompressed_size, int flag)
 {
     unsigned char *ZF = malloc(16);
 
+    /* Intimate friendship with this variable in filters/zisofs.c */
+    extern int iso_zisofs2_enable_susp_z2;
+
     if (ZF == NULL) {
         return ISO_OUT_OF_MEM;
     }
     ZF[0] = 'Z';
     ZF[1] = 'F';
     ZF[2] = (unsigned char) 16;
-    ZF[3] = (unsigned char) 1;
-    ZF[4] = (unsigned char) 'p';
-    ZF[5] = (unsigned char) 'z';
+    if (algo[0] == 'p' && algo[1] == 'z') {
+        ZF[3] = (unsigned char) 1;
+    } else {
+        ZF[3] = (unsigned char) 2;
+        if (iso_zisofs2_enable_susp_z2)
+            ZF[1] = '2';
+    }
+    ZF[4] = (unsigned char) algo[0];
+    ZF[5] = (unsigned char) algo[1];
     ZF[6] = (unsigned char) header_size_div4;
     ZF[7] = (unsigned char) block_size_log2;
-    iso_bb(&ZF[8], uncompressed_size, 4);
+    if (algo[0] == 'p' && algo[1] == 'z') {
+        if (uncompressed_size > (uint64_t) 0xffffffff)
+            return ISO_ZISOFS_TOO_LARGE;
+        iso_bb(&ZF[8], (uint32_t) uncompressed_size, 4);
+    } else {
+        iso_lsb64(&ZF[8], uncompressed_size);
+    }
     if (to_ce) {
         return susp_append_ce(t, susp, ZF);
     } else {
@@ -830,25 +998,31 @@ int zisofs_add_ZF(Ecma119Image *t, struct susp_info *susp, int to_ce,
 
 
 /* @param flag bit0= Do not add data but only count sua_free and ce_len
+               bit1= account for crossing block boundaries
+                     (implied by bit0 == 0)
 */
 static
 int add_zf_field(Ecma119Image *t, Ecma119Node *n, struct susp_info *info,
-                 size_t *sua_free, size_t *ce_len, int flag)
+                 size_t *sua_free, size_t *ce_len, size_t base_ce, int flag)
 {
     int ret, will_copy = 1, stream_type = 0, do_zf = 0;
     int header_size_div4 = 0, block_size_log2 = 0;
-    uint32_t uncompressed_size = 0;
+    uint64_t uncompressed_size = 0;
     IsoStream *stream = NULL, *input_stream, *last_stream, *first_stream;
     IsoStream *first_filter = NULL;
     IsoFile *file;
     void *xipt;
     struct zisofs_zf_info *zf;
+    uint8_t algo[2];
 
     /* Intimate friendship with this function in filters/zisofs.c */
     int ziso_is_zisofs_stream(IsoStream *stream, int *stream_type,
+                              uint8_t zisofs_algo[2],
                               int *header_size_div4, int *block_size_log2,
-                              uint32_t *uncompressed_size, int flag);
+                              uint64_t *uncompressed_size, int flag);
 
+    if (!(flag & 1))
+        flag |= 2;
 
     if (iso_node_get_type(n->node) != LIBISO_FILE)
         return 2;
@@ -887,7 +1061,8 @@ int add_zf_field(Ecma119Image *t, Ecma119Node *n, struct susp_info *info,
     }
 
     /* Determine stream type : 1=ziso , -1=osiz , 0=other */
-    ret = ziso_is_zisofs_stream(stream, &stream_type, &header_size_div4,
+    algo[0] = algo[1] = 0;
+    ret = ziso_is_zisofs_stream(stream, &stream_type, algo, &header_size_div4,
                                 &block_size_log2, &uncompressed_size, 0);
     if (ret < 0)
         return ret;
@@ -898,7 +1073,7 @@ int add_zf_field(Ecma119Image *t, Ecma119Node *n, struct susp_info *info,
         do_zf = 1;
     } else if(first_stream == last_stream || !will_copy) {
         /* Try whether the image side stream remembers a ZF field */
-        ret = iso_stream_get_src_zf(first_stream, &header_size_div4,
+        ret = iso_stream_get_src_zf(first_stream, algo, &header_size_div4,
                                     &block_size_log2, &uncompressed_size, 0);
         if (ret == 1 && header_size_div4 > 0)
             do_zf = 1;
@@ -912,6 +1087,8 @@ int add_zf_field(Ecma119Image *t, Ecma119Node *n, struct susp_info *info,
             header_size_div4 = zf->header_size_div4;
             block_size_log2 = zf->block_size_log2;
             uncompressed_size = zf->uncompressed_size;
+            algo[0] = zf->zisofs_algo[0];
+            algo[1] = zf->zisofs_algo[1];
             if (header_size_div4 > 0)
                 do_zf = 1;
         }
@@ -922,7 +1099,7 @@ int add_zf_field(Ecma119Image *t, Ecma119Node *n, struct susp_info *info,
 
     /* Account for field size */
     if (*sua_free < 16 || *ce_len > 0) {
-        *ce_len += 16;
+        susp_calc_add_to_ce(ce_len, base_ce, 16, flag & 2);
     } else {
         *sua_free -= 16;
     }
@@ -930,7 +1107,7 @@ int add_zf_field(Ecma119Image *t, Ecma119Node *n, struct susp_info *info,
         return 1;
 
     /* write ZF field */
-    ret = zisofs_add_ZF(t, info, (*ce_len > 0), header_size_div4,
+    ret = zisofs_add_ZF(t, info, (*ce_len > 0), algo, header_size_div4,
                        block_size_log2, uncompressed_size, 0);
     if (ret < 0)
         return ret;
@@ -970,10 +1147,15 @@ int aaip_xinfo_cloner(void *old_data, void **new_data, int flag)
  * a CE entry of 28 bytes in SUA, this computation fails if not the 28 bytes
  * are taken into account at start. In this case the caller should retry with
  * bit0 set. 
+ * If the resulting *ce added to base_ce is in a different block than base_ce,
+ * then computation with bit0 fails and the caller should finally try bit1.
  * 
  * @param flag      bit0= assume CA usage (else return 0 on SUA overflow)
+ *                  bit1= let CA start at block start (else return 0 if
+ *                        *ce crosses a block boundary)
  * @return          1= ok, computation of *su_size and *ce is valid
  *                  0= not ok, CA usage is necessary but bit0 was not set
+ *                             or *ce crosses boundary and bit1 was not set
  *                     (*su_size and *ce stay unaltered in this case)
  *                 <0= error:
  *                 -1= not enough SUA space for 28 bytes of CE entry
@@ -981,27 +1163,57 @@ int aaip_xinfo_cloner(void *old_data, void **new_data, int flag)
  */
 static
 int susp_calc_nm_sl_al(Ecma119Image *t, Ecma119Node *n, size_t space,
-                       size_t *su_size, size_t *ce, int flag)
+                       size_t *su_size, size_t *ce, size_t base_ce, int flag)
 {
     char *name;
-    size_t namelen, su_mem, ce_mem;
+    size_t namelen, su_mem, ce_mem, ce_prepad = 0;
     void *xipt;
     size_t num_aapt = 0, sua_free = 0;
     int ret;
+    uint8_t *aapt;
+
+#ifdef Libisofs_ce_calc_debug_extrA
+
+    if (n->node->name != NULL)
+        fprintf(stderr, "libburn_DEBUG: susp_calc_nm_sl_al : %.f %s \n",
+                        (double) base_ce, n->node->name);
+
+#endif /* Libisofs_ce_calc_debug_extrA */
 
     su_mem = *su_size;
     ce_mem = *ce;
     if (*ce > 0 && !(flag & 1))
         goto unannounced_ca;
 
+    if (flag & 2) {
+        flag |= 1;
+        if (base_ce % BLOCK_SIZE) {
+
+#ifdef Libisofs_ce_calc_debuG
+
+            fprintf(stderr,
+                "\nlibburn_DEBUG: Accounting for %d bytes CE padding : %s\n\n",
+                (int) (BLOCK_SIZE - (base_ce % BLOCK_SIZE)), n->node->name);
+
+#endif /* Libisofs_ce_calc_debuG */
+
+            ce_prepad = BLOCK_SIZE - (base_ce % BLOCK_SIZE);
+        }
+    }
+
+    namelen = 0;
     name = get_rr_fname(t, n->node->name);
-    namelen = strlen(name);
-    free(name);
+    if (name != NULL) {
+        namelen = strlen(name);
+        free(name);
+    }
 
     if (flag & 1) {
        /* Account for 28 bytes of CE field */
-       if (*su_size + 28 > space)
+       if (*su_size + 28 > space) {
+           *ce += ce_prepad;
            return -1;
+       }
        *su_size += 28;
     }
 
@@ -1020,7 +1232,7 @@ int susp_calc_nm_sl_al(Ecma119Image *t, Ecma119Node *n, size_t space,
                 of the name will always fit into the directory entry.)
         */;
 
-        *ce = 5 + namelen;
+        susp_calc_add_to_ce(ce, base_ce, 5 + namelen, flag & 2);
         *su_size = space;
     }
     if (n->type == ECMA119_SYMLINK) {
@@ -1032,8 +1244,10 @@ int susp_calc_nm_sl_al(Ecma119Image *t, Ecma119Node *n, size_t space,
         int cew = (*ce != 0); /* are we writing to CA ? */
 
         dest = get_rr_fname(t, ((IsoSymlink*)n->node)->dest);
-        if (dest == NULL)
+        if (dest == NULL) {
+            *ce += ce_prepad;
             return -2;
+        }
         prev = dest;
         cur = strchr(prev, '/');
         while (1) {
@@ -1060,7 +1274,7 @@ int susp_calc_nm_sl_al(Ecma119Image *t, Ecma119Node *n, size_t space,
                     /* 
                      * ok, we need a Continuation Area anyway
                      * TODO this can be handled better, but for now SL
-                     * will be completelly moved into the CA
+                     * will be completely moved into the CA
                      */
                     if (!(flag & 1)) {
                         free(dest);
@@ -1090,21 +1304,26 @@ int susp_calc_nm_sl_al(Ecma119Image *t, Ecma119Node *n, size_t space,
                              * the component can be divided between this
                              * and another SL entry
                              */
-                            *ce += 255; /* this SL, full */
-                            sl_len = 5 + (clen - fit);
+                            /* Will fill up old SL and write it */
+                            susp_calc_add_to_ce(ce, base_ce, 255, flag & 2);
+                            sl_len = 5 + (clen - fit); /* Start new SL */
                         } else {
                             /*
                              * the component will need a 2rd SL entry in
                              * any case, so we prefer to don't write 
                              * anything in this SL
                              */
-                            *ce += sl_len + 255;
-                            sl_len = 5 + (clen - 250) + 2;
+                            /* Will write non-full old SL */
+                            susp_calc_add_to_ce(ce, base_ce, sl_len, flag & 2);
+                            /* Will write another full SL */
+                            susp_calc_add_to_ce(ce, base_ce, 255, flag & 2);
+                            sl_len = 5 + (clen - 250) + 2; /* Start new SL */
                         }
                     } else {
                         /* case 2, create a new SL entry */
-                        *ce += sl_len;
-                        sl_len = 5 + clen;
+                        /* Will write non-full old SL */
+                        susp_calc_add_to_ce(ce, base_ce, sl_len, flag & 2);
+                        sl_len = 5 + clen; /* Start new SL */
                     }
                 } else {
                     sl_len += clen;
@@ -1126,18 +1345,21 @@ int susp_calc_nm_sl_al(Ecma119Image *t, Ecma119Node *n, size_t space,
             /* the whole SL fits into the SUA */
             *su_size += sl_len;
         } else {
-            *ce += sl_len;
+            susp_calc_add_to_ce(ce, base_ce, sl_len, flag & 2);
         }
 
     }
 
     /* Find out whether ZF is to be added and account for its bytes */
     sua_free = space - *su_size;
-    add_zf_field(t, n, NULL, &sua_free, ce, 1);
+    add_zf_field(t, n, NULL, &sua_free, ce, base_ce, 1 | (flag & 2));
     *su_size = space - sua_free;
     if (*ce > 0 && !(flag & 1))
         goto unannounced_ca;
 
+    *ce += ce_prepad;
+    ce_prepad = 0;
+
     /* obtain num_aapt from node */
     xipt = NULL;
     num_aapt = 0;
@@ -1150,12 +1372,47 @@ int susp_calc_nm_sl_al(Ecma119Image *t, Ecma119Node *n, size_t space,
     /* let the expert decide where to add num_aapt */
     if (num_aapt > 0) {
         sua_free = space - *su_size;
-        aaip_add_AL(t, NULL, NULL, num_aapt, &sua_free, ce, 1);
+        aapt = (uint8_t *) xipt;
+        aaip_add_AL(t, NULL, &aapt, num_aapt, &sua_free, ce, base_ce,
+                    1 | (flag & 2));
         *su_size = space - sua_free;
         if (*ce > 0 && !(flag & 1))
             goto unannounced_ca;
     }
 
+#ifdef Libisofs_ce_calc_debug_filetraP
+
+    if (n->node->name != NULL)
+        if (strcmp(n->node->name, "...insert.leaf.name.here...") == 0)
+            fprintf(stderr,
+                    "libburn_DEBUG: filename breakpoint susp_calc_nm_sl_al\n");
+
+#endif /* Libisofs_ce_calc_debug_filetraP */
+
+    if (*ce > 0 && !(flag & 2)) {
+        if (base_ce / BLOCK_SIZE !=
+            (base_ce + *ce + ISO_CE_ENTRY_SIZE - 1) / BLOCK_SIZE) {
+
+#ifdef Libisofs_ce_calc_debuG
+
+            fprintf(stderr,
+    "\nlibburn_DEBUG: Crossed block boundary: %.f (%lu) -> %.f (%lu) : %s\n\n",
+                    (double) base_ce, (unsigned long) (base_ce / BLOCK_SIZE),
+                    (double) (base_ce + *ce - 1),
+                    (unsigned long)
+                        ((base_ce + *ce + ISO_CE_ENTRY_SIZE - 1) / BLOCK_SIZE),
+                     n->node->name);
+
+#endif /* Libisofs_ce_calc_debuG */
+
+            /* Crossed a block boundary */
+            *su_size = su_mem;
+            *ce = ce_mem;
+            return 0;
+        }
+    }
+
+    *ce += ce_prepad;
     return 1;
 
 unannounced_ca:;
@@ -1167,10 +1424,12 @@ unannounced_ca:;
 
 /* @param flag bit0= Do not add data but only count sua_free and ce_len
                      param info may be NULL in this case
+               bit1= account for crossing block boundaries
+                     (implied by bit0 == 0)
 */
 static
 int add_aa_string(Ecma119Image *t, Ecma119Node *n, struct susp_info *info,
-                  size_t *sua_free, size_t *ce_len, int flag)
+                  size_t *sua_free, size_t *ce_len, size_t base_ce, int flag)
 {
     int ret;
     uint8_t *aapt;
@@ -1185,18 +1444,20 @@ int add_aa_string(Ecma119Image *t, Ecma119Node *n, struct susp_info *info,
         num_aapt = aaip_count_bytes((unsigned char *) xipt, 0);
         if (num_aapt > 0) {
             if (flag & 1) {
-                ret = aaip_add_AL(t, NULL,NULL, num_aapt, sua_free, ce_len, 1);
+                aapt = (unsigned char *) xipt;
+                ret = aaip_add_AL(t, NULL, &aapt, num_aapt, sua_free, ce_len,
+                                  base_ce, flag & (1 | 2));
             } else {
                 aapt = malloc(num_aapt);
                 if (aapt == NULL)
                     return ISO_OUT_OF_MEM;
                 memcpy(aapt, xipt, num_aapt);
                 ret = aaip_add_AL(t, info, &aapt, num_aapt, sua_free, ce_len,
-                                  0);
+                                  base_ce, 0);
+                /* aapt is NULL now and the memory is owned by t */
             }
             if (ret < 0) 
                 return ret;
-            /* aapt is NULL now and the memory is owned by t */
         }
     }
     return 1;
@@ -1214,11 +1475,13 @@ int add_aa_string(Ecma119Image *t, Ecma119Node *n, struct susp_info *info,
  *      Already occupied space in the directory record.
  * @param ce
  *      Will be filled with the space needed in a CE
+ * @param base_ce
+ *      Predicted fill of continuation area by previous nodes of same dir
  * @return
  *      The size needed for the RR entries in the System Use Area
  */
 size_t rrip_calc_len(Ecma119Image *t, Ecma119Node *n, int type, size_t used_up,
-                     size_t *ce)
+                     size_t *ce, size_t base_ce)
 {
     size_t su_size, space;
     int ret;
@@ -1281,9 +1544,11 @@ size_t rrip_calc_len(Ecma119Image *t, Ecma119Node *n, int type, size_t used_up,
     if (type == 0) {
 
         /* Try without CE */
-        ret = susp_calc_nm_sl_al(t, n, space, &su_size, ce, 0);
-        if (ret == 0) /* Retry with CE */
-            ret = susp_calc_nm_sl_al(t, n, space, &su_size, ce, 1);
+        ret = susp_calc_nm_sl_al(t, n, space, &su_size, ce, base_ce, 0);
+        if (ret == 0) /* Retry with CE but no block crossing */
+            ret = susp_calc_nm_sl_al(t, n, space, &su_size, ce, base_ce, 1);
+        if (ret == 0) /* Retry with aligned CE and block hopping */
+            ret = susp_calc_nm_sl_al(t, n, space, &su_size, ce, base_ce, 1 | 2);
         if (ret == -2)
            return ISO_OUT_OF_MEM;
 
@@ -1310,9 +1575,13 @@ size_t rrip_calc_len(Ecma119Image *t, Ecma119Node *n, int type, size_t used_up,
             if (t->opts->aaip && !t->opts->aaip_susp_1_10) {
                 *ce += 160; /* ER of AAIP */
             }
-            /* Compute length of AAIP string of root node */
+            /* Compute length of AAIP string of root node.
+               Will write all AIIP to CA, which already starts at
+               block boundary. So no need for three tries.
+             */
             aaip_sua_free= 0;
-            ret = add_aa_string(t, n, NULL, &aaip_sua_free, &aaip_len, 1);
+            ret = add_aa_string(t, n, NULL, &aaip_sua_free, &aaip_len, base_ce,
+                                1 | 2);
             if (ret < 0)
                return ret;
             *ce += aaip_len;
@@ -1360,7 +1629,7 @@ void susp_info_free(struct susp_info* susp)
  *      Pointer to the struct susp_info where the entries will be stored.
  *      If some entries need to go to a Continuation Area, they will be added
  *      to the existing ce_susp_fields, and ce_len will be incremented
- *      propertly. Please ensure ce_block is initialized propertly.
+ *      properly. Please ensure ce_block is initialized properly.
  * @return
  *      1 success, < 0 error
  */
@@ -1376,7 +1645,7 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
     size_t rrip_er_len= 182;
     size_t su_size_pd, ce_len_pd; /* predicted sizes of SUA and CA */
     int ce_is_predicted = 0;
-    size_t aaip_sua_free= 0, aaip_len= 0;
+    size_t aaip_sua_free= 0, aaip_len= 0, ce_mem;
     int space;
 
     if (t == NULL || n == NULL || info == NULL) {
@@ -1393,6 +1662,18 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
         return ISO_ASSERT_FAILURE;
     }
 
+    /* Mark start index of node's continuation area for later update */
+    info->current_ce_start = info->n_ce_susp_fields;
+    ce_mem = info->ce_len;
+
+#ifdef Libisofs_ce_calc_debug_filetraP
+
+    if (n->node->name != NULL)
+        if (strcmp(n->node->name, "...put.leafname.here...") == 0)
+            fprintf(stderr, "libburn_DEBUG: filename breakpoint\n");
+
+#endif /* Libisofs_ce_calc_debug_filetraP */
+
     if (type == 2 && n->parent != NULL) {
         node = n->parent;
     } else {
@@ -1493,19 +1774,29 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
         uint8_t **comps= NULL; /* components of the SL field */
         size_t n_comp = 0; /* number of components */
 
+        namelen = 0;
         name = get_rr_fname(t, n->node->name);
+        if (name == NULL)
+            name = strdup("");
+        if (name == NULL) {
+            ret = ISO_OUT_OF_MEM;
+            goto add_susp_cleanup;
+        }
         namelen = strlen(name);
-
         sua_free = space - info->suf_len;
 
         /* Try whether NM, SL, AL will fit into SUA */
         su_size_pd = info->suf_len;
         ce_len_pd = ce_len;
         ret = susp_calc_nm_sl_al(t, n, (size_t) space,
-                                 &su_size_pd, &ce_len_pd, 0);
+                                 &su_size_pd, &ce_len_pd, info->ce_len, 0);
         if (ret == 0) { /* Have to use CA. 28 bytes of CE are necessary */
             ret = susp_calc_nm_sl_al(t, n, (size_t) space,
-                                     &su_size_pd, &ce_len_pd, 1);
+                                     &su_size_pd, &ce_len_pd, info->ce_len, 1);
+            if (ret == 0) /* Retry with aligned CE and block hopping */
+                ret = susp_calc_nm_sl_al(t, n, (size_t) space,
+                                         &su_size_pd, &ce_len_pd, info->ce_len,
+                                         1 | 2);
             sua_free -= 28;
             ce_is_predicted = 1;
         }
@@ -1535,6 +1826,13 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
             int cew = (nm_type == 1); /* are we writing to CE? */
 
             dest = get_rr_fname(t, ((IsoSymlink*)n->node)->dest);
+            if (dest == NULL)
+                dest = strdup("");
+            if (dest == NULL) {
+                ret = ISO_OUT_OF_MEM;
+                goto add_susp_cleanup;
+            }
+
             prev = dest;
             cur = strchr(prev, '/');
             while (1) {
@@ -1568,7 +1866,7 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
                         /* 
                          * ok, we need a Continuation Area anyway
                          * TODO this can be handled better, but for now SL
-                         * will be completelly moved into the CA
+                         * will be completely moved into the CA
                          */
 
                         /* sua_free, ce_len, nm_type already account for CE */
@@ -1699,6 +1997,26 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
         }
 
         if (ce_is_predicted) {
+            if ((info->ce_len % BLOCK_SIZE) &&
+                (info->ce_len + ce_len_pd - 1 ) / BLOCK_SIZE !=
+                info->ce_len / BLOCK_SIZE) {
+                /* Linux fs/isofs wants byte_offset + ce_len <= BLOCK_SIZE
+                 * Insert padding to shift CE offset to next block start
+                 */
+
+#ifdef Libisofs_ce_calc_debuG
+
+                fprintf(stderr,
+                  "\nlibburn_DEBUG: Inserting %d bytes of CE padding : %s\n\n",
+                  (int) (BLOCK_SIZE - (info->ce_len % BLOCK_SIZE)),
+                   n->node->name);
+
+#endif /* Libisofs_ce_calc_debuG */
+
+                ret = pseudo_susp_add_PAD(t, info);
+                if (ret < 0)
+                    goto add_susp_cleanup;
+            }
             /* Add the CE entry */
             ret = susp_add_CE(t, ce_len_pd, info);
             if (ret < 0) {
@@ -1740,14 +2058,14 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
         }
 
         /* Eventually write zisofs ZF field */
-        ret = add_zf_field(t, n, info, &sua_free, &ce_len, 0);
+        ret = add_zf_field(t, n, info, &sua_free, &ce_len, ce_mem, 0);
         if (ret < 0)
             goto add_susp_cleanup;
 
         /* Eventually obtain AAIP field string from node
            and write it to directory entry or CE area.
         */
-        ret = add_aa_string(t, n, info, &sua_free, &ce_len, 0);
+        ret = add_aa_string(t, n, info, &sua_free, &ce_len, ce_mem, 0);
         if (ret < 0)
             goto add_susp_cleanup;
 
@@ -1808,7 +2126,8 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
 
             /* Compute length of AAIP string of root node */
             aaip_sua_free= 0;
-            ret = add_aa_string(t, n, NULL, &aaip_sua_free, &aaip_len, 1);
+            ret = add_aa_string(t, n, NULL, &aaip_sua_free, &aaip_len, ce_mem,
+                                1 | 2);
             if (ret < 0)
                 goto add_susp_cleanup;
 
@@ -1829,7 +2148,8 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
             }
             /* Write AAIP string of root node */
             aaip_sua_free= aaip_len= 0;
-            ret = add_aa_string(t, n, info, &aaip_sua_free, &aaip_len, 0);
+            ret = add_aa_string(t, n, info, &aaip_sua_free, &aaip_len, ce_mem,
+                                0);
             if (ret < 0)
                 goto add_susp_cleanup;
 
@@ -1848,29 +2168,85 @@ int rrip_get_susp_fields(Ecma119Image *t, Ecma119Node *n, int type,
     return ISO_SUCCESS;
 
     add_susp_cleanup: ;
-    free(name);
-    free(dest);
+    if (name != NULL)
+        free(name);
+    if (dest != NULL)
+        free(dest);
     susp_info_free(info);
     return ret;
 }
 
+
+/* Update the sizes of CE fields at end of info->susp_fields and in
+   single node range of info->ce_susp_fields.
+*/
+static
+int susp_update_CE_sizes(Ecma119Image *t, struct susp_info *info, int flag)
+{
+    size_t i, curr_pos;
+    uint8_t *curr_ce;
+    uint32_t size;
+
+    if (info->n_susp_fields == 0 ||
+        info->n_ce_susp_fields - info->current_ce_start == 0)
+        return ISO_SUCCESS;
+
+    for (i = 0; i < info->n_susp_fields; i++)
+        if (info->susp_fields[i][0] == 'C')
+            if(info->susp_fields[i][1] == 'E')
+    break;
+    if (i >= info->n_susp_fields) {
+        iso_msg_submit(t->image->id, ISO_ASSERT_FAILURE, 0,
+                       "System Use Area field contains no CE, but there are fields in Continuation Area");
+        return ISO_ASSERT_FAILURE;
+    }
+    curr_ce = info->susp_fields[i];
+    curr_pos = 0;
+    for (i = info->current_ce_start; i < info->n_ce_susp_fields; i++) {
+       if (info->ce_susp_fields[i][0] == 0) {
+           curr_pos = 0; /* pseudo SUSP PAD */
+    continue;
+       }
+       if (info->ce_susp_fields[i][0] == 'C' &&
+           info->ce_susp_fields[i][1] == 'E') {
+          size = (curr_pos + info->ce_susp_fields[i][2]) % BLOCK_SIZE;
+          if (size == 0)
+              size = BLOCK_SIZE;
+          iso_bb(curr_ce + 20, size, 4);
+          curr_ce = info->ce_susp_fields[i];
+       }
+       curr_pos = (curr_pos + info->ce_susp_fields[i][2]) % 2048;
+    }
+    if (curr_pos > 0) {
+        size = curr_pos % BLOCK_SIZE;
+        iso_bb(curr_ce + 20, size, 4);
+    }
+    return ISO_SUCCESS;
+}
+
+
 /**
  * Write the given SUSP fields into buf. Note that Continuation Area
  * fields are not written.
  * If info does not contain any SUSP entry this function just return. 
  * After written, the info susp_fields array will be freed, and the counters
- * updated propertly.
+ * updated properly.
  */
 void rrip_write_susp_fields(Ecma119Image *t, struct susp_info *info,
                             uint8_t *buf)
 {
     size_t i;
     size_t pos = 0;
+    int ret;
 
     if (info->n_susp_fields == 0) {
         return;
     }
 
+    ret = susp_update_CE_sizes(t, info, 0);
+    if (ret < 0)
+        return;
+
     for (i = 0; i < info->n_susp_fields; i++) {
         memcpy(buf + pos, info->susp_fields[i], info->susp_fields[i][2]);
         pos += info->susp_fields[i][2];
@@ -1896,6 +2272,7 @@ int rrip_write_ce_fields(Ecma119Image *t, struct susp_info *info)
     size_t i;
     uint8_t *padding = NULL;
     int ret= ISO_SUCCESS;
+    uint64_t written = 0, pad_size;
 
     if (info->n_ce_susp_fields == 0) {
         goto ex;
@@ -1903,11 +2280,24 @@ int rrip_write_ce_fields(Ecma119Image *t, struct susp_info *info)
     LIBISO_ALLOC_MEM(padding, uint8_t, BLOCK_SIZE);
 
     for (i = 0; i < info->n_ce_susp_fields; i++) {
+        if (info->ce_susp_fields[i][0] == 0) {
+            /* Pseudo field: pad up to next block boundary */
+            pad_size = BLOCK_SIZE - (written % BLOCK_SIZE);
+            if (pad_size == BLOCK_SIZE)
+    continue;
+            memset(padding, 0, pad_size);
+            ret = iso_write(t, padding, pad_size);
+            if (ret < 0)
+                goto write_ce_field_cleanup;
+            written += pad_size;
+    continue;
+        }
         ret = iso_write(t, info->ce_susp_fields[i], 
                         info->ce_susp_fields[i][2]);
         if (ret < 0) {
             goto write_ce_field_cleanup;
         }
+        written += info->ce_susp_fields[i][2];
     }
 
     /* pad continuation area until block size */
@@ -1915,6 +2305,9 @@ int rrip_write_ce_fields(Ecma119Image *t, struct susp_info *info)
     if (i > 0 && i < BLOCK_SIZE) {
         memset(padding, 0, i);
         ret = iso_write(t, padding, i);
+        if (ret < 0)
+            goto write_ce_field_cleanup;
+        written += i;
     }
 
     write_ce_field_cleanup: ;
@@ -1925,6 +2318,7 @@ int rrip_write_ce_fields(Ecma119Image *t, struct susp_info *info)
     free(info->ce_susp_fields);
     info->ce_susp_fields = NULL;
     info->n_ce_susp_fields = 0;
+    info->alloc_ce_susp_fields = 0;
     info->ce_len = 0;
 ex:;
     LIBISO_FREE_MEM(padding);
