@@ -1,16 +1,22 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
- * Copyright (c) 2009 Thomas Schmitt
+ * Copyright (c) 2009 - 2015 Thomas Schmitt
  *
  * This file is part of the libisofs project; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License version 2 as
- * published by the Free Software Foundation. See COPYING file for details.
+ * modify it under the terms of the GNU General Public License version 2 
+ * or later as published by the Free Software Foundation. 
+ * See COPYING file for details.
  */
 
+#ifdef HAVE_CONFIG_H
+#include "../config.h"
+#endif
+
 #include "libisofs.h"
 #include "stream.h"
 #include "fsource.h"
 #include "util.h"
+#include "node.h"
 
 #include <stdlib.h>
 #include <string.h>
@@ -152,8 +158,56 @@ int fsrc_update_size(IsoStream *stream)
     return ISO_SUCCESS;
 }
 
+static 
+IsoStream *fsrc_get_input_stream(IsoStream *stream, int flag)
+{
+    return NULL;
+}
+
+int fsrc_clone_stream(IsoStream *old_stream, IsoStream **new_stream,
+                      int flag)
+{
+    FSrcStreamData *data, *new_data;
+    IsoStream *stream;
+    int ret;
+
+    if (flag)
+        return ISO_STREAM_NO_CLONE; /* unknown option required */
+
+    data = (FSrcStreamData*) old_stream->data;
+    if (data->src->class->version < 2)
+        return ISO_STREAM_NO_CLONE; /* No clone_src() method available */
+
+    *new_stream = NULL;
+    stream = calloc(1, sizeof(IsoStream));
+    if (stream == NULL)
+        return ISO_OUT_OF_MEM;
+    new_data = calloc(1, sizeof(FSrcStreamData));
+    if (new_data == NULL) {
+        free((char *) stream);
+        return ISO_OUT_OF_MEM;
+    }
+    *new_stream = stream;
+    stream->class = old_stream->class;
+    stream->refcount = 1;
+    stream->data = new_data;
+
+    ret = data->src->class->clone_src(data->src, &(new_data->src), 0);
+    if (ret < 0) {
+        free((char *) stream);
+        free((char *) new_data);
+        return ret;
+    }
+    new_data->dev_id = data->dev_id;
+    new_data->ino_id = data->ino_id;
+    new_data->size = data->size;
+
+    return ISO_SUCCESS;
+}
+
+static
 IsoStreamIface fsrc_stream_class = {
-    1, /* update_size is defined for this stream */
+    4, /* version */
     "fsrc",
     fsrc_open,
     fsrc_close,
@@ -162,7 +216,10 @@ IsoStreamIface fsrc_stream_class = {
     fsrc_is_repeatable,
     fsrc_get_id,
     fsrc_free,
-    fsrc_update_size
+    fsrc_update_size,
+    fsrc_get_input_stream,
+    NULL,
+    fsrc_clone_stream
 };
 
 int iso_file_source_stream_new(IsoFileSource *src, IsoStream **stream)
@@ -233,25 +290,25 @@ int iso_file_source_stream_new(IsoFileSource *src, IsoStream **stream)
 }
 
 
-int iso_stream_get_src_zf(IsoStream *stream, int *header_size_div4,
-                          int *block_size_log2, uint32_t *uncompressed_size,
-                          int flag)
+int iso_stream_get_src_zf(IsoStream *stream, uint8_t zisofs_algo[2],
+                          int *header_size_div4, int *block_size_log2,
+                          uint64_t *uncompressed_size, int flag)
 {
     int ret;
     FSrcStreamData *data;
     IsoFileSource *src;
 
     /* Intimate friendship with libisofs/fs_image.c */
-    int iso_ifs_source_get_zf(IsoFileSource *src, int *header_size_div4,
-                  int *block_size_log2, uint32_t *uncompressed_size, int flag);
+    int iso_ifs_source_get_zf(IsoFileSource *src, uint8_t zisofs_algo[2],
+                              int *header_size_div4, int *block_size_log2,
+                              uint64_t *uncompressed_size, int flag);
 
     if (stream->class != &fsrc_stream_class)
         return 0;
     data = stream->data;
     src = data->src;
-
-    ret = iso_ifs_source_get_zf(src, header_size_div4, block_size_log2,
-                                uncompressed_size, 0);
+    ret = iso_ifs_source_get_zf(src, zisofs_algo, header_size_div4,
+                                block_size_log2, uncompressed_size, 0);
     return ret;
 }
 
@@ -333,7 +390,7 @@ static
 int cut_out_read(IsoStream *stream, void *buf, size_t count)
 {
     struct cut_out_stream *data = stream->data;
-    count = (size_t)MIN(data->size - data->pos, count);
+    count = (size_t) MIN((size_t) (data->size - data->pos), count);
     if (count == 0) {
         return 0;
     }
@@ -370,11 +427,68 @@ void cut_out_free(IsoStream *stream)
     free(data);
 }
 
+static
+int cut_out_update_size(IsoStream *stream)
+{
+    return ISO_SUCCESS;
+}
+
+static 
+IsoStream* cut_out_get_input_stream(IsoStream *stream, int flag)
+{
+    return NULL;
+}
+
+static
+int cut_out_clone_stream(IsoStream *old_stream, IsoStream **new_stream,
+                      int flag)
+{
+    struct cut_out_stream *data, *new_data;
+    IsoStream *stream;
+    int ret;
+
+    if (flag)
+        return ISO_STREAM_NO_CLONE; /* unknown option required */
+
+    data = (struct cut_out_stream *) old_stream->data;
+    if (data->src->class->version < 2)
+        return ISO_STREAM_NO_CLONE; /* No clone_src() method available */
+
+    *new_stream = NULL;
+    stream = calloc(1, sizeof(IsoStream));
+    if (stream == NULL)
+        return ISO_OUT_OF_MEM;
+    stream->refcount = 1;
+    stream->class = old_stream->class;
+    new_data = calloc(1, sizeof(struct cut_out_stream));
+    if (new_data == NULL) {
+        free((char *) stream);
+        return ISO_OUT_OF_MEM;
+    }
+    ret = data->src->class->clone_src(data->src, &(new_data->src), 0);
+    if (ret < 0) {
+        free((char *) stream);
+        free((char *) new_data);
+        return ret;
+    }
+
+    new_data->dev_id = (dev_t) 0;
+    new_data->ino_id = cut_out_serial_id++;
+    new_data->offset = data->offset;
+    new_data->size = data->size;
+    new_data->pos = 0;
+
+    stream->data = new_data;
+    *new_stream = stream;
+    return ISO_SUCCESS;
+}
+
 /*
  * TODO update cut out streams to deal with update_size(). Seems hard.
  */
+static
 IsoStreamIface cut_out_stream_class = {
-    0,
+    4, /* version */
     "cout",
     cut_out_open,
     cut_out_close,
@@ -382,7 +496,12 @@ IsoStreamIface cut_out_stream_class = {
     cut_out_read,
     cut_out_is_repeatable,
     cut_out_get_id,
-    cut_out_free
+    cut_out_free,
+    cut_out_update_size,
+    cut_out_get_input_stream,
+    NULL,
+    cut_out_clone_stream
+    
 };
 
 int iso_cut_out_stream_new(IsoFileSource *src, off_t offset, off_t size,
@@ -512,7 +631,7 @@ int mem_read(IsoStream *stream, void *buf, size_t count)
         return ISO_FILE_NOT_OPENED;
     }
 
-    if (data->offset >= data->size) {
+    if (data->offset >= (ssize_t) data->size) {
         return 0; /* EOF */
     }
 
@@ -544,12 +663,69 @@ void mem_free(IsoStream *stream)
 {
     MemStreamData *data;
     data = (MemStreamData*)stream->data;
-    free(data->buf);
+    if (data->buf != NULL)
+        free(data->buf);
     free(data);
 }
 
+static
+int mem_update_size(IsoStream *stream)
+{
+    return ISO_SUCCESS;
+}
+
+static 
+IsoStream* mem_get_input_stream(IsoStream *stream, int flag)
+{
+    return NULL;
+}
+
+static
+int mem_clone_stream(IsoStream *old_stream, IsoStream **new_stream,
+                      int flag)
+{
+    MemStreamData *data, *new_data;
+    IsoStream *stream;
+    uint8_t *new_buf = NULL;
+
+    if (flag)
+        return ISO_STREAM_NO_CLONE; /* unknown option required */
+
+    *new_stream = NULL;
+    stream = calloc(1, sizeof(IsoStream));
+    if (stream == NULL)
+        return ISO_OUT_OF_MEM;
+    stream->refcount = 1;
+    stream->class = old_stream->class;
+    new_data = calloc(1, sizeof(MemStreamData));
+    if (new_data == NULL) {
+        free((char *) stream);
+        return ISO_OUT_OF_MEM;
+    }
+    data = (MemStreamData *) old_stream->data;
+    if (data->size > 0) {
+        new_buf = calloc(1, data->size);
+        if (new_buf == NULL) {
+            free((char *) stream);
+            free((char *) new_data);
+            return ISO_OUT_OF_MEM;
+        }
+        memcpy(new_buf, data->buf, data->size);
+    }
+    new_data->buf = new_buf;
+    new_data->offset = -1;
+    new_data->ino_id = mem_serial_id++;
+    new_data->size = data->size;
+
+    stream->data = new_data;
+    *new_stream = stream;
+    return ISO_SUCCESS;
+}
+
+
+static
 IsoStreamIface mem_stream_class = {
-    0,
+    4, /* version */
     "mem ",
     mem_open,
     mem_close,
@@ -557,7 +733,12 @@ IsoStreamIface mem_stream_class = {
     mem_read,
     mem_is_repeatable,
     mem_get_id,
-    mem_free
+    mem_free,
+    mem_update_size,
+    mem_get_input_stream,
+    NULL,
+    mem_clone_stream
+
 };
 
 /**
@@ -565,7 +746,7 @@ IsoStreamIface mem_stream_class = {
  * When the Stream refcount reach 0, the buffer is free(3).
  *
  * @return
- *      1 sucess, < 0 error
+ *      1 success, < 0 error
  */
 int iso_memory_stream_new(unsigned char *buf, size_t size, IsoStream **stream)
 {
@@ -581,7 +762,7 @@ int iso_memory_stream_new(unsigned char *buf, size_t size, IsoStream **stream)
         return ISO_OUT_OF_MEM;
     }
     data = malloc(sizeof(MemStreamData));
-    if (str == NULL) {
+    if (data == NULL) {
         free(str);
         return ISO_OUT_OF_MEM;
     }
@@ -664,7 +845,12 @@ void iso_stream_get_file_name(IsoStream *stream, char *name)
     if (!strncmp(type, "fsrc", 4)) {
         FSrcStreamData *data = stream->data;
         char *path = iso_file_source_get_path(data->src);
-        strncpy(name, path, PATH_MAX);
+        if (path == NULL) {
+            name[0] = 0;
+            return;
+        }
+        strncpy(name, path, PATH_MAX - 1);
+        name[PATH_MAX - 1] = 0;
         free(path);
     } else if (!strncmp(type, "boot", 4)) {
         strcpy(name, "BOOT CATALOG");
@@ -677,17 +863,27 @@ void iso_stream_get_file_name(IsoStream *stream, char *name)
     }
 }
 
+/* @param flag  bit0= Obtain most fundamental stream */
 IsoStream *iso_stream_get_input_stream(IsoStream *stream, int flag)
 {
     IsoStreamIface* class;
+    IsoStream *result = NULL, *next;
 
     if (stream == NULL) {
         return NULL;
     }
-    class = stream->class;
-    if (class->version < 2)
-        return NULL;
-    return class->get_input_stream(stream, 0);
+    while (1) {
+        class = stream->class;
+        if (class->version < 2)
+            return result;
+        next = class->get_input_stream(stream, 0);
+        if (next == NULL)
+            return result;
+        result = next;
+        if (!(flag & 1))
+            return result;
+        stream = result;
+    }
 }
 
 char *iso_stream_get_source_path(IsoStream *stream, int flag)
@@ -720,12 +916,22 @@ ex:;
     return path;
 }
 
-/* @return 1 = ok , 0 = not an ISO image stream , <0 = error */
+/*
+   @param flag bit0= in case of filter stream do not dig for base stream
+   @return 1 = ok , 0 = not an ISO image stream , <0 = error
+*/
 int iso_stream_set_image_ino(IsoStream *stream, ino_t ino, int flag)
 {
+    IsoStream *base_stream;
+
     if (stream == NULL) {
         return ISO_NULL_POINTER;
     }
+    if (!(flag & 1)) {
+        base_stream = iso_stream_get_input_stream(stream, 1);
+        if (base_stream != NULL)
+            stream = base_stream;
+    }
     if (stream->class == &fsrc_stream_class) {
         FSrcStreamData *fsrc_data = stream->data;
         fsrc_data->ino_id = ino;
@@ -734,6 +940,104 @@ int iso_stream_set_image_ino(IsoStream *stream, ino_t ino, int flag)
    return 0;
 }
 
+int iso_stream_cmp_ifs_sections(IsoStream *s1, IsoStream *s2, int *cmp_ret,
+                                int flag)
+{   
+    int ret;
+    FSrcStreamData *fssd1, *fssd2;
+    IsoFileSource *src1, *src2;
+
+    /* Must keep any suspect in the game to preserve transitivity of the
+       calling function by ranking applicable streams lower than
+       non-applicable. ones.
+    */
+    if (s1->class != &fsrc_stream_class && s2->class != &fsrc_stream_class)
+        return 0;
+
+    /* Compare eventual image data section LBA and sizes */
+    if (s1->class == &fsrc_stream_class) {
+        fssd1= (FSrcStreamData *) s1->data;
+        src1 = fssd1->src;
+    } else {
+        src1 = NULL;
+    }
+    if (s2->class == &fsrc_stream_class) {
+        fssd2= (FSrcStreamData *) s2->data;
+        src2 = fssd2->src;
+    } else {
+        src2 = NULL;
+    }
+    ret = iso_ifs_sections_cmp(src1, src2, cmp_ret, 1);
+    if (ret <= 0)
+        return 0;
+    return 1;
+}
+
+
+/* Maintain and exploit a list of stream compare functions seen by
+   iso_stream_cmp_ino(). This is needed to separate stream comparison
+   families in order to keep iso_stream_cmp_ino() transitive while
+   alternative stream->class->cmp_ino() decide inside the families.
+*/
+struct iso_streamcmprank {
+    int (*cmp_func)(IsoStream *s1, IsoStream *s2);
+    struct iso_streamcmprank *next;
+};
+
+static struct iso_streamcmprank *streamcmpranks = NULL;
+
+static
+int iso_get_streamcmprank(int (*cmp_func)(IsoStream *s1, IsoStream *s2),
+                          int flag)
+{
+    int idx;
+    struct iso_streamcmprank *cpr, *last_cpr = NULL;
+
+    idx = 0;
+    for (cpr = streamcmpranks; cpr != NULL; cpr = cpr->next) {
+        if (cpr->cmp_func == cmp_func)
+    break;
+        idx++;
+        last_cpr = cpr;
+    }
+    if (cpr != NULL)
+        return idx;
+    LIBISO_ALLOC_MEM_VOID(cpr, struct iso_streamcmprank, 1);
+    cpr->cmp_func = cmp_func;
+    cpr->next = NULL;
+    if (last_cpr != NULL)
+        last_cpr->next = cpr;
+    if (streamcmpranks == NULL)
+        streamcmpranks = cpr;
+    return idx;
+ex:;
+    return -1;
+}
+
+static
+int iso_cmp_streamcmpranks(int (*cf1)(IsoStream *s1, IsoStream *s2),
+                           int (*cf2)(IsoStream *s1, IsoStream *s2))
+{
+    int rank1, rank2;
+
+    rank1 = iso_get_streamcmprank(cf1, 0);
+    rank2 = iso_get_streamcmprank(cf2, 0);
+    return rank1 < rank2 ? -1 : 1;
+}
+
+int iso_stream_destroy_cmpranks(int flag)
+{
+    struct iso_streamcmprank *cpr, *next;
+
+    for (cpr = streamcmpranks; cpr != NULL; cpr = next) {
+        next = cpr->next;
+        LIBISO_FREE_MEM(cpr);
+    }
+    streamcmpranks = NULL;
+    return ISO_SUCCESS;
+}
+
+
 /* API */ 
 int iso_stream_cmp_ino(IsoStream *s1, IsoStream *s2, int flag)
 {
@@ -742,8 +1046,6 @@ int iso_stream_cmp_ino(IsoStream *s1, IsoStream *s2, int flag)
     dev_t dev_id1, dev_id2;
     ino_t ino_id1, ino_id2;
     off_t size1, size2;
-    FSrcStreamData *fssd1, *fssd2;
-
 
 /*
    #define Libisofs_stream_cmp_ino_debuG 1
@@ -760,10 +1062,72 @@ int iso_stream_cmp_ino(IsoStream *s1, IsoStream *s2, int flag)
     if (s2 == NULL)
         return 1;
 
-    if (s1->class->version >= 3 && !(flag & 1)) {
-       /* Filters may have smarter methods to compare themselves with others */
-       ret = s1->class->cmp_ino(s1, s2);
-       return ret;
+    /* This stays transitive by the fact that 
+       iso_stream_cmp_ifs_sections() is transitive,
+       returns > 0 if s1 or s2 are applicable,
+       ret is -1 if s1 is applicable but s2 is not,
+       ret is 1 if s1 is not applicable but s2 is.
+
+       Proof:
+       Be A the set of applicable streams, S and G transitive and
+       antisymmetric relations in respect to outcome {-1, 0, 1}.
+       The combined relation R shall be defined by
+         I.   R(a,b) = S(a,b) if a in A or b in A, else G(a,b)
+       Further S shall have the property
+         II.  S(a,b) = -1 if a in A and b not in A
+       Then R can be proven to be transitive:
+       By enumerating the 8 combinations of a,b,c being in A or not, we get
+       5 cases of pure S or pure G. Three cases are mixed:
+         a,b not in A, c in A : G(a,b) == -1, S(b,c) == -1 -> S(a,c) == -1 
+             Impossible because S(b,c) == -1 contradicts II.
+         a,c not in A, b in A : S(a,b) == -1, S(b,c) == -1 -> G(a,c) == -1
+             Impossible because S(a,b) == -1 contradicts II.
+         b,c not in A, a in A : S(a,b) == -1, G(b,c) == -1 -> S(a,c) == -1
+             Always true because S(a,c) == -1 by definition II.
+    */
+    if (iso_stream_cmp_ifs_sections(s1, s2, &ret, 0) > 0)
+        return ret; /* Both are unfiltered from loaded ISO filesystem */
+
+    if (!(flag & 1)) {
+       /* Filters may have smarter methods to compare themselves with others.
+          Transitivity is ensured by ranking mixed pairs by the rank of their
+          comparison functions, and by ranking streams with .cmp_ino lower
+          than streams without.
+          (One could merge (class->version < 3) and (cmp_ino == NULL).)
+
+          Here we define S for "and" rather than "or"
+            I.   R(a,b) = S(a,b) if a in A and b in A, else G(a,b)
+          and the function ranking in case of "exor" makes sure that
+            II.  G(a,b) = -1 if a in A and b not in A
+          Again we get three mixed cases:
+            a not in A, b,c in A : G(a,b) == -1, S(b,c) == -1 -> G(a,c) == -1 
+                Impossible because G(a,b) == -1 contradicts II.
+            b not in A, a,c in A : G(a,b) == -1, G(b,c) == -1 -> S(a,c) == -1
+                Impossible because G(b,c) == -1 contradicts II.
+            c not in A, a,b in A : S(a,b) == -1, G(b,c) == -1 -> G(a,c) == -1
+                Always true because G(a,c) == -1 by definition II.
+       */
+       if ((s1->class->version >= 3) ^ (s2->class->version >= 3)) {
+           /* One of both has no own com_ino function. Rank it as larger. */
+           return s1->class->version >= 3 ? -1 : 1;
+       } else if (s1->class->version >= 3) {
+           if (s1->class->cmp_ino == s2->class->cmp_ino) {
+               if (s1->class->cmp_ino == NULL) {
+                   /* Both are NULL. No decision by .cmp_ino(). */;
+               } else {
+                   /* Both are compared by the same function */
+                   ret = s1->class->cmp_ino(s1, s2);
+                   return ret;
+               }
+           } else {
+               /* Not the same cmp_ino() function. Decide by list rank of
+                  function while building the list on the fly.
+               */
+               ret = iso_cmp_streamcmpranks(s1->class->cmp_ino,
+                                            s2->class->cmp_ino);
+               return ret;
+           }
+       }
     }
 
     iso_stream_get_id(s1, &fs_id1, &dev_id1, &ino_id1);
@@ -819,14 +1183,6 @@ int iso_stream_cmp_ino(IsoStream *s1, IsoStream *s2, int flag)
 
     if (s1->class != s2->class)
         return (s1->class < s2->class ? -1 : 1);
-    if (s1->class == &fsrc_stream_class) {
-        /* Compare eventual image data section LBA and sizes */
-        fssd1= (FSrcStreamData *) s1->data;
-        fssd2= (FSrcStreamData *) s2->data;
-        ret = iso_ifs_sections_cmp(fssd1->src, fssd2->src, 0);
-        if (ret != 0)
-            return ret;
-    }
     if (fs_id1 == 0 && dev_id1 == 0 && ino_id1 == 0) {
         return (s1 < s2 ? -1 : 1);
     }
@@ -863,9 +1219,6 @@ int iso_stream_read_buffer(IsoStream *stream, char *buf, size_t count,
     return 1;
 }
 
-#ifdef Libisofs_with_checksumS
-
-
 /* @param flag bit0= dig out most original stream (e.g. because from old image)
    @return 1=ok, md5 is valid,
            0= not ok, 
@@ -873,14 +1226,15 @@ int iso_stream_read_buffer(IsoStream *stream, char *buf, size_t count,
 */  
 int iso_stream_make_md5(IsoStream *stream, char md5[16], int flag)
 {
-    int res, is_open = 0;
-    char buffer[2048];
+    int ret, is_open = 0;
+    char * buffer = NULL;
     void *ctx= NULL;
     off_t file_size;
     uint32_t b, nblocks;
     size_t got_bytes;
     IsoStream *input_stream;
 
+    LIBISO_ALLOC_MEM(buffer, char, 2048);
     if (flag & 1) {
         while(1) {
            input_stream = iso_stream_get_input_stream(stream, 0);
@@ -891,37 +1245,74 @@ int iso_stream_make_md5(IsoStream *stream, char md5[16], int flag)
     }
 
     if (! iso_stream_is_repeatable(stream))
-        return 0;
-    res = iso_md5_start(&ctx);
-    if (res < 0)
-        return res;
-    res = iso_stream_open(stream);
-    if (res < 0)
-        return 0;
+        {ret = 0; goto ex;}
+    ret = iso_md5_start(&ctx);
+    if (ret < 0)
+        goto ex;
+    ret = iso_stream_open(stream);
+    if (ret < 0)
+        goto ex;
     is_open = 1;
     file_size = iso_stream_get_size(stream);
     nblocks = DIV_UP(file_size, 2048);
     for (b = 0; b < nblocks; ++b) {
-        res = iso_stream_read_buffer(stream, buffer, 2048, &got_bytes);
-        if (res < 0) {
-            res = 0;
+        ret = iso_stream_read_buffer(stream, buffer, 2048, &got_bytes);
+        if (ret < 0) {
+            ret = 0;
             goto ex;
         }
         /* Do not use got_bytes to stay closer to IsoFileSrc processing */
         if (file_size - b * 2048 > 2048)
-            res = 2048;
+            ret = 2048;
         else
-            res = file_size - b * 2048;
-        iso_md5_compute(ctx, buffer, res);
+            ret = file_size - b * 2048;
+        iso_md5_compute(ctx, buffer, ret);
     }
-    res = 1;
+    ret = 1;
 ex:;
     if (is_open)
         iso_stream_close(stream);
     if (ctx != NULL)
         iso_md5_end(&ctx, md5);
-    return res;
+    LIBISO_FREE_MEM(buffer);
+    return ret;
 }
 
-#endif /* Libisofs_with_checksumS */
+/* API */
+int iso_stream_clone(IsoStream *old_stream, IsoStream **new_stream, int flag)
+{
+    int ret;
+
+    if (old_stream->class->version < 4)
+        return ISO_STREAM_NO_CLONE;
+    ret = old_stream->class->clone_stream(old_stream, new_stream, 0);
+    return ret;
+}
+
+int iso_stream_clone_filter_common(IsoStream *old_stream,
+                                   IsoStream **new_stream,
+                                   IsoStream **new_input, int flag)
+{
+    IsoStream *stream, *input_stream;
+    int ret;
+
+    *new_stream = NULL;
+    *new_input = NULL;
+    input_stream = iso_stream_get_input_stream(old_stream, 0);
+    if (input_stream == NULL)
+        return ISO_STREAM_NO_CLONE;
+    stream = calloc(1, sizeof(IsoStream));
+    if (stream == NULL)
+        return ISO_OUT_OF_MEM;
+    ret = iso_stream_clone(input_stream, new_input, 0);
+    if (ret < 0) {
+        free((char *) stream);
+        return ret;
+    }
+    stream->class = old_stream->class;
+    stream->refcount = 1;
+    stream->data = NULL;
+    *new_stream = stream;
+    return ISO_SUCCESS;
+}
 
