@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
+ * Copyright (c) 2014 Thomas Schmitt
  * 
  * This file is part of the libisofs project; you can redistribute it and/or 
  * modify it under the terms of the GNU General Public License version 2 
@@ -322,9 +323,10 @@ int iso_htable_create(size_t size, hash_funtion_t hash,
 {
     IsoHTable *t;
     
-    if (table == NULL) {
-        return ISO_OUT_OF_MEM;
-    }
+    if (size <= 0)
+        return ISO_WRONG_ARG_VALUE;
+    if (table == NULL)
+        return ISO_NULL_POINTER;
     
     t = malloc(sizeof(IsoHTable));
     if (t == NULL) {
